package yukihane.inqubus.manager;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import saccubus.worker.WorkerListener;
import saccubus.worker.impl.MessageReportable;
import saccubus.worker.impl.PercentageReportable;
import saccubus.worker.impl.convert.Convert;
import saccubus.worker.impl.convert.ConvertProgress;
import saccubus.worker.impl.convert.ConvertResult;
import saccubus.worker.impl.download.Download;
import saccubus.worker.impl.download.DownloadProgress;
import saccubus.worker.impl.download.DownloadResult;
import saccubus.worker.profile.ConvertProfile;
import saccubus.worker.profile.DownloadProfile;

/**
 *
 * @author yuki
 */
public class TaskManage {

    private static final Logger logger = Logger.getLogger(TaskManage.class.getName());
    private final ExecutorService downloadExecutorService;
    private final ExecutorService convertExecutorService;
    private final Map<Integer, ManageTarget<DownloadResult>> downloadTargets = new HashMap<>();
    private final Map<Integer, ManageTarget<ConvertResult>> convertTargets = new HashMap<>();
    private final TaskManageListener clientListener;

    public TaskManage() {
        this(1, 1, null);
    }

    public TaskManage(int maxDownload, int maxConvert) {
        this(maxDownload, maxConvert, null);
    }

    public TaskManage(int maxDownload, int maxConvert, TaskManageListener listener) {
        downloadExecutorService = Executors.newFixedThreadPool(maxDownload);
        convertExecutorService = Executors.newFixedThreadPool(maxConvert);
        this.clientListener = listener;
    }

    public synchronized boolean add(RequestProcess request) {
        final DownloadProfile dp = request.getDownloadProfile();
        final ConvertProfile cp = request.getConvertProfile();
        if (dp != null && (dp.getVideoProfile().isDownload() || dp.getCommentProfile().isDownload())) {
            // ダウンロードするものがあればまずダウンロード処理
            final Download task = new Download(dp, request.getVideoId(),
                    new DownloadListener(request.getRowId()));
            final Future<DownloadResult> future = downloadExecutorService.submit(task);
            downloadTargets.put(request.getRowId(), new ManageTarget<>(request, future));
            return true;

        } else if (cp != null && cp.isConvert()) {
            final Convert task = new Convert(cp, dp.getVideoProfile().getLocalFile(), dp.getCommentProfile().
                    getLocalFile(), new ConvertListener(request.getRowId()));
            final Future<ConvertResult> future = convertExecutorService.submit(task);
            convertTargets.put(request.getRowId(), new ManageTarget<>(request, future));
            return true;
        }
        return false;
    }

    private class DownloadListener extends TaskManageInnerListener<DownloadResult, DownloadProgress> {

        private DownloadListener(int rowId) {
            super(rowId);
        }

        @Override
        public void done(DownloadResult result) {
            super.done(result);
            // TODO 変換が必要なら変換キューに入れる
        }

        @Override
        protected TaskKind getKind() {
            return TaskKind.DOWNLOAD;
        }
    }

    private class ConvertListener extends TaskManageInnerListener<ConvertResult, ConvertProgress> {

        private ConvertListener(int rowId) {
            super(rowId);
        }

        @Override
        protected TaskKind getKind() {
            return TaskKind.CONVERT;
        }
    }

    abstract class TaskManageInnerListener<T, V extends PercentageReportable & MessageReportable> implements WorkerListener<T, V> {

        private final int rowId;

        protected TaskManageInnerListener(int rowId) {
            this.rowId = rowId;
        }

        private void notify(TaskStatus status) {
            notify(status, 0.0, "");
        }

        private void notify(TaskStatus status, double percentage, String message) {
            if (getListener() == null) {
                return;
            }
            getListener().process(rowId, getKind(), status, percentage, message);
        }

        private TaskManageListener getListener() {
            return clientListener;
        }

        protected abstract TaskKind getKind();

        @Override
        public void process(V progress) {
            logger.log(Level.FINEST, "process: {0}", progress);
            notify(TaskStatus.DOING, progress.getPercentage(), progress.getMessage());
        }

        @Override
        public void cancelled() {
            logger.log(Level.FINE, "cancelled: {0}", toString());
            notify(TaskStatus.CANCELLED);
        }

        @Override
        public void done(T result) {
            logger.log(Level.FINE, "done: {0}", result);
            notify(TaskStatus.DONE);
        }

        @Override
        public void error(Throwable th) {
            logger.log(Level.SEVERE, "error", th);
            notify(TaskStatus.ERROR, 0.0, th.getMessage());
        }
    }

    class ManageTarget<T> {

        private final RequestProcess request;
        private final Future<T> future;

        ManageTarget(RequestProcess request, Future<T> future) {
            this.request = request;
            this.future = future;
        }

        Future<T> getFuture() {
            return future;
        }

        RequestProcess getRequest() {
            return request;
        }
    }
}